@@ -257,7 +257,15 @@ read_metadata (SoupMessage *msg, gboolean to_blank)
 		if (got_lf) {
 			if (!to_blank)
 				break;
-			if (nread == 1 || (nread == 2 && read_buf[0] == '\r'))
+			if (nread == 1 &&
+			    !strncmp ((char *)io->read_meta_buf->data +
+				      io->read_meta_buf->len - 2,
+				      "\n\n", 2))
+				break;
+			else if (nread == 2 &&
+				 !strncmp ((char *)io->read_meta_buf->data +
+					   io->read_meta_buf->len - 3,
+					   "\n\r\n", 3))
 				break;
 		}
 	}
